from stringology.align import align, mismatches
from stringology.text import remove_accents


class SpSim:
    """
    SpSim is a spelling similarity measure for identifying cognates by
    learning cross-language spelling differences.

    >>> sim = SpSim()
    >>> sim('phase', 'fase')
    0.6

    Learning contextualized spelling differences from an example:

    >>> sim.learn([('alpha', 'alfa')])
    >>> sorted(sim.diffs.items())
    [('ph\\tf', 'la')]

    SpSim has learned that 'ph' may be replaced by 'f' if 'ph' comes after 'l'
    and before 'a'.

    Generalizing the admissible contexts of known spelling differences:

    >>> sim.learn([('phase', 'fase')])
    >>> sorted(sim.diffs.items())
    [('ph\\tf', '*a')]

    SpSim has learned that 'ph' may be replaced by 'f' if it comes before 'a'.

    >>> sim.learn([('photo', 'foto')])
    >>> sorted(sim.diffs.items())
    [('ph\\tf', '**')]

    SpSim has learned that 'ph' may be replaced by 'f'.

    >>> sim('phenomenal', 'fenomenal')
    1.0

    """

    def __init__(self,
                 examples=None,
                 ignore_case=True,
                 ignore_accents=True,
                 group_vowels=False,
                 no_empty=False):
        self.ignore_case = ignore_case
        self.ignore_accents = ignore_accents
        self.group_vowels = group_vowels
        self.no_empty = no_empty
        self.diffs = {}
        if examples:
            self.learn(examples)

    def __call__(self, a, b, known=None, unknown=None):
        dist = self._dist(a, b, known=known, unknown=unknown)
        return 1.0 - dist / max(1, len(a), len(b))

    def _dist(self, a, b, known=None, unknown=None):
        a, b = self._prepare(a, b)
        dist = 0  # total distance
        for nchars, diff, ctxt in self._get_diffs(a, b):
            if not SpSim._match_context(ctxt, self.diffs.get(diff, None)):
                dist += nchars
                if unknown is not None:
                    unknown.append((diff, ctxt))
            elif known is not None:
                known.append((diff, ctxt))
        return dist

    def learn(self, examples, trace_fn=None):
        for a, b in examples:
            a, b = self._prepare(a, b)
            for _, diff, ctxt in self._get_diffs(a, b):
                before = self.diffs.get(diff, None)
                after = SpSim._generalize_context(before, ctxt)
                self.diffs[diff] = after
                if trace_fn is not None:
                    trace_fn(a, b, diff, before, after)

    def _prepare(self, a, b):
        if self.ignore_case:
            a, b = a.lower(), b.lower()
        if self.ignore_accents:
            a, b = remove_accents(a), remove_accents(b)
        return a, b

    def _get_diffs(self, a, b):
        alignment = align(" ^" + a + "$ ", " ^" + b + "$ ", gap=" ")
        for mma, mmb in mismatches(*alignment, context=2):
            nchars = len(mma) - 4  # discount the left and right context chars
            diffa = mma[2:-2].replace(" ", "")
            diffb = mmb[2:-2].replace(" ", "")
            ctxtl = mma[1]
            ctxtr = mma[-2]
            if self.no_empty and (not diffa or not diffb):
                if diffa.startswith(ctxtl) or diffb.startswith(ctxtl):
                    diffa = mma[1:-2].replace(" ", "")
                    diffb = mmb[1:-2].replace(" ", "")
                    ctxtl = mma[0]
                elif diffa.endswith(ctxtr) or diffb.endswith(ctxtr):
                    diffa = mma[2:-1].replace(" ", "")
                    diffb = mmb[2:-1].replace(" ", "")
                    ctxtr = mma[-1]
                else:
                    diffa = mma[1:-1].replace(" ", "")
                    diffb = mmb[1:-1].replace(" ", "")
                    ctxtl = mma[0]
                    ctxtr = mma[-1]
            ctxtl = self._get_context_repr(ctxtl)
            ctxtr = self._get_context_repr(ctxtr)
            yield nchars, diffa + "\t" + diffb, ctxtl + ctxtr

    def _get_context_repr(self, char):
        # TODO: add vowels for other alphabets
        if self.group_vowels and char in "aeiou":
            return "a"  # vowel representative
        return char

    @staticmethod
    def _match_context(ctxt, learned):
        return (
            learned and
            learned[0] in "*" + ctxt[0] and
            learned[1] in "*" + ctxt[1]
        )

    @staticmethod
    def _generalize_context(learned, ctxt):
        if not learned:
            return ctxt
        lft = "*" if learned[0] != ctxt[0] else learned[0]
        rgt = "*" if learned[1] != ctxt[1] else learned[1]
        return lft + rgt



import itertools

import munkres
from spsim.spsim import SpSim


class PhraseSpSim(SpSim):
    '''
    Measures the similarity between pairs of phrases using SpSim for comparing
    words.

    A phrase is a sequence one or more words.

                January 2010, Luís Gomes <luismsgomes@gmail.com>

    Works as follows:
    1 - Compute the distance between each word A of one phrase and each word B
        of the  other phrase using (1 - SpSim(A, B)) * max(len(A), len(B)).
    2 - Compute the one-to-one assignment of words from one phrase to words
        of the other phrase that minimizes the sum of all pair distances. If
        the phrases have a different number of words, then we add null words
        to the shortest phrase. The distance between any word C and null is
        len(C).
    3 - The similarity score is computed as 1 - D/L, where D is the minimum
        total distance (computed in 2) and L is the length of the longest
        phrase, excluding whitespace.

    Step 2 is computed using the Munkres algorithm (aka Hungarian). Note that
    the problem solved in step 2 is widely known as "the assignment problem".

    >>> sim = PhraseSpSim()
    >>> sim.learn([('photo', 'foto'), ('alpha', 'alfa'), ('pangea', 'pangeia')])
    >>> sorted(sim.diffs.items())
    [('\\ti', 'ea'), ('ph\\tf', '**')]

    SpSim has learned that 'ph' may be replaced by 'f' and 'ea' by 'eia'.

    >>> sim('phenomenal idea', 'ideia fenomenal')
    1.0

    '''
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __call__(self, a, b, known=None, unknown=None):
        # TODO: update known and unknown lists, when given
        if isinstance(a, str):
            a = a.split()
        if isinstance(b, str):
            b = b.split()
        if not isinstance(a, (list, tuple)) or not isinstance(b, (list, tuple)):
            raise ValueError
        if a == b:
            return 1.0
        if not a or not b:
            return 0.0
        if len(a) == 1 or len(b) == 1:
            if len(a) == len(b):  # == 1
                sim = SpSim.__call__(self, a[0], b[0])
            else:
                sim = self._match_compound(a, b)
                if sim < 1.0:
                    sim = max(sim, self._match_words(a, b))
        else:
            sim = self._match_words(a, b)
        return sim

    def _match_compound(self, a, b):
        assert len(a) == 1 or len(b) == 1
        if len(a) == 1:
            return max(
                SpSim.__call__(self, a[0], ''.join(bp))
                for bp in itertools.permutations(b)
            )
        return max(
            SpSim.__call__(self, ''.join(ap), b[0])
            for ap in itertools.permutations(a)
        )

    def _match_words(self, a, b):
        # compute the length of the longest phrase (excluding spaces)
        L = max(sum(map(len, a)), sum(map(len, b)))
        # let's obtain the (square) distance matrix
        if len(a) > len(b):
            b.extend(['']*(len(a)-len(b)))
        elif len(a) < len(b):
            a.extend(['']*(len(b)-len(a)))
        M = [[self._dist(p, q) for q in b] for p in a]
        # now compute the one to one assignment between a and b that
        # minimizes the sum of distances (this is "the assignment problem")
        D = sum(M[row][col] for row, col in munkres.Munkres().compute(M))
        return 1.0 - D / L



from setuptools import setup, find_packages
from os import path
import re


def packagefile(*relpath):
    return path.join(path.dirname(__file__), *relpath)


def read(*relpath):
    with open(packagefile(*relpath)) as f:
        return f.read()


def get_version(*relpath):
    match = re.search(
        r'''^__version__ = ['"]([^'"]*)['"]''',
        read(*relpath),
        re.M
    )
    if not match:
        raise RuntimeError('Unable to find version string.')
    return match.group(1)


setup(
    name='spsim',
    version=get_version('src', 'spsim', '__init__.py'),
    description='A spelling similarity measure for cognate identification.',
    long_description=read('README.rst'),
    url='https://github.com/luismsgomes/spsim',
    author='Luís Gomes',
    author_email='luismsgomes@gmail.com',
    license='MIT',
    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.5',
        'Topic :: Text Processing :: Linguistic',
    ],
    keywords='text bilingual cognate mt machine translation',
    setup_requires=[
        'pytest',
        'coverage',
        'pytest-cov',
    ],
    install_requires=[
        'docopt',
        'openfile',
        'munkres',
        'stringology',
    ],
    packages=find_packages('src'),
    package_dir={'': 'src'},
    entry_points={
        'console_scripts': [
            'spsim=spsim.__main__:main',
        ],
    },
)



"""
Usage: spsim [options] <examples-file> [<input-file> [<output-file>]]

Options
    --trace <fname>
                Save learned differences into specified file.
    --phrases   Assume that input contains phrases instead of words.
    --case      Be case sensitive.
    --accents   Be sensitive to accents.
    --debug     Output two additional columns with known and unknown
                differences found on each word pair.
                Warning: this option will not work correctly with --phrases.
    --vowels    Group vowels.
    --no-empty  Ensure differences are not empty on any side.
"""

import logging
import docopt
import openfile
from spsim import PhraseSpSim, SpSim


LOGGER = logging.getLogger("spsim")


def read_examples(fname):
    with (openfile.openfile(fname)) as lines:
        for line in lines:
            columns = line.rstrip("\n").split("\t")
            if len(columns) >= 2:
                yield columns[:2]


def main():
    logging.basicConfig()
    opts = docopt.docopt(__doc__)
    cls = PhraseSpSim if opts["--phrases"] else SpSim
    sim = cls(
        ignore_case=not opts["--case"],
        ignore_accents=not opts["--accents"],
        group_vowels=opts["--vowels"],
        no_empty=opts["--no-empty"],
    )
    examples = read_examples(opts["<examples-file>"])
    if opts["--trace"]:
        with (openfile.openfile(opts["--trace"], "wt")) as out:
            def trace(a, b, diff, before, after):
                diffa, diffb = diff.split("\t")
                after_repr = "%s[%s]%s <=> %s[%s]%s" % (
                    after[0], diffa, after[1],
                    after[0], diffb, after[1],
                )
                if before and before != after:
                    before_repr = "%s[%s]%s <=> %s[%s]%s" % (
                        before[0], diffa, before[1],
                        before[0], diffb, before[1],
                    )
                    out.write("%s<=>%s => %s => %s\n" % (
                        a, b, before_repr, after_repr
                    ))
                else:
                    out.write("%s<=>%s => %s\n" % (a, b, after_repr))

            sim.learn(examples, trace_fn=trace)
    else:
        sim.learn(examples)
    inputfile = openfile.openfile(opts["<input-file>"])
    outputfile = openfile.openfile(opts["<output-file>"], "wt")
    debug = opts["--debug"]
    with inputfile, outputfile:
        for num, line in enumerate(inputfile, start=1):
            cols = line.rstrip('\n').split('\t')
            if len(cols) < 2:
                LOGGER.warning(
                    'line %d of %s has less than 2 columns; skipping',
                    num, opts["<input-file>"]
                )
                continue

            if debug:
                known, unknown = [], []
                cols.append(sim(cols[0], cols[1], known=known, unknown=unknown))
                cols.append(" ".join(d.replace("\t", "<=>") for d in known))
                cols.append(" ".join(d.replace("\t", "<?>") for d in unknown))
            else:
                cols.append(sim(cols[0], cols[1]))
            print(*cols, sep='\t', file=outputfile)


if __name__ == '__main__':
    main()
